from flask import Blueprint
from flask import jsonify
from flask_login import current_user
from flask_login import login_required
from sqlalchemy import and_
from sqlalchemy.exc import SQLAlchemyError

from asreview.project import ASReviewProject
from asreview.webapp import DB
from asreview.webapp.authentication.models import Project
from asreview.webapp.authentication.models import User

bp = Blueprint("team", __name__, url_prefix="/api")

REQUESTER_FRAUD = {"message": "Request can not made by current user."}


@bp.route("/projects/<project_id>/users", methods=["GET"])
@login_required
def users(project_id):
    """Returns all users involved in a project."""
    response = jsonify(REQUESTER_FRAUD), 404

    # get project
    project = Project.query.filter(Project.project_id == project_id).one_or_none()

    # check if this project is in fact from current user
    if project in current_user.projects:
        # get associated users from project
        collaborators = project.collaborators
        invitations = project.pending_invitations

        # get all users that are involved (invited or collabo)
        collaborators = [user.id for user in collaborators]
        invitations = [user.id for user in invitations]

        # get all users minus myself
        all_users = [
            u.summarize()
            for u in User.query.filter(and_(User.public, User.id != current_user.id))
            .order_by("name")
            .all()
        ]

        # response
        response = (
            jsonify(
                {
                    "all_users": all_users,
                    "collaborators": collaborators,
                    "invitations": invitations,
                }
            ),
            200,
        )
    return response


@bp.route("/projects/<project_id>/users/<user_id>", methods=["DELETE"])
@login_required
def end_collaboration(project_id, user_id):
    """Project owner removes a collaborator, or collaborator
    removes him/herself."""
    response = jsonify(REQUESTER_FRAUD), 404
    # get project
    project = Project.query.filter(Project.project_id == project_id).one_or_none()

    # check if project is owned by current user or if the user is
    # involved in the project
    if project and (
        (project.owner == current_user) or (project in current_user.involved_in)
    ):
        user = DB.session.get(User, user_id)

        try:
            project.collaborators.remove(user)
            DB.session.commit()
            response = (
                jsonify({"message": "Collaborator removed from project."}),
                200
            )

        except SQLAlchemyError:
            response = (
                jsonify({"message": "Error removing collaborator."}),
                404
            )
    return response


@bp.route("/invitations", methods=["GET"])
@login_required
def pending_invitations():
    """Returns pending invitations for current user."""
    invitations = []
    for p in current_user.pending_invitations:
        # get path of project
        path = p.project_path
        # get object to get name
        asreview_object = ASReviewProject(path)
        # append info
        invitations.append(
            {
                "id": p.id,
                "project_id": p.project_id,
                "owner_id": p.owner_id,
                "name": asreview_object.config["name"],
                "created_at_unix": asreview_object.config["created_at_unix"],
                "mode": asreview_object.config["mode"],
            }
        )
    response = (jsonify({"invited_for_projects": invitations}), 200)
    return response


@bp.route("/invitations/projects/<project_id>/users/<user_id>", methods=["POST"])
@login_required
def invite(project_id, user_id):
    """Project owner invites a user to collaborate on a project"""
    response = jsonify(REQUESTER_FRAUD), 404
    # get project
    project = Project.query.filter(Project.project_id == project_id).one_or_none()
    # check if project is from current user
    if project and project.owner == current_user:
        user = DB.session.get(User, user_id)
        project.pending_invitations.append(user)
        try:
            DB.session.commit()
            response = (
                jsonify({"message": f'User "{user.identifier}" invited.'}),
                200
            )
        except SQLAlchemyError:
            response = (
                jsonify({"message": f'User "{user.identifier}" not invited.'}),
                404,
            )
    return response


@bp.route("/invitations/projects/<project_id>/accept", methods=["POST"])
@login_required
def accept_invitation(project_id):
    """Invited person accepts an invitation."""
    response = jsonify(REQUESTER_FRAUD), 404
    # get project
    project = Project.query.filter(Project.project_id == project_id).one_or_none()
    # if user is current user, try to add this user to project
    if project and current_user in project.pending_invitations:
        # remove invitation
        project.pending_invitations.remove(current_user)
        # add as collaborator
        project.collaborators.append(current_user)
        try:
            DB.session.commit()
            response = (
                jsonify({"message": "User accepted invitation for project."}),
                200
            )
        except SQLAlchemyError:
            response = (
                jsonify({"message": "Error accepting invitation."}),
                404
            )
    return response


@bp.route("/invitations/projects/<project_id>/reject", methods=["DELETE"])
@login_required
def reject_invitation(project_id):
    """Invited person rejects an invitation."""
    response = jsonify(REQUESTER_FRAUD), 404
    # get project
    project = Project.query.filter(Project.project_id == project_id).one_or_none()
    # if current_user is indeed invited
    if project and current_user in project.pending_invitations:
        # remove invitation
        project.pending_invitations.remove(current_user)
        try:
            DB.session.commit()
            response = (
                jsonify({"message": "User rejected invitation for project."}),
                200
            )
        except SQLAlchemyError:
            response = (
                jsonify({"message": "Error rejecting invitation."}),
                404
            )
    return response


@bp.route("/invitations/projects/<project_id>/users/<user_id>", methods=["DELETE"])
@login_required
def delete_invitation(project_id, user_id):
    """removes an invitation"""
    response = jsonify(REQUESTER_FRAUD), 404
    # get project
    project = Project.query.filter(Project.project_id == project_id).one_or_none()
    # check if project is from current user
    if project and project.owner == current_user:
        # get user
        user = DB.session.get(User, user_id)
        # remove from project
        project.pending_invitations.remove(user)
        try:
            DB.session.commit()
            response = jsonify({"message": "Owner deleted invitation."}), 200
        except SQLAlchemyError:
            response = jsonify({"message": "Error deleting invitation."}), 404
    return response



import configparser
from pathlib import Path

from asreview.webapp.authentication.models import User

config_file = "asreview.ini"
config_dir = "config"

config = configparser.ConfigParser()
BASE_DIR = Path(__file__).resolve().parent.parent

CONFIG_FILE = BASE_DIR.joinpath(config_dir).joinpath(config_file)

config.read(CONFIG_FILE)


# get user (1 of 3)
def get_user(test_user_id):
    """Returns a User model based on a test user
    account that can be found in the config file.
    The test_user_id refers to the position of the
    user account credentials in the .ini file
    (1, 2, or 3)"""
    section = config[f"user{test_user_id}"]
    # create user
    user = User(
        section["email"],
        email=section["email"],
        name=section["name"],
        affiliation=section["affiliation"],
        password=section["password"],
    )
    # store password
    user.password = section["password"]
    return user


def get_user_data(test_user_id):
    """Returns the data for a user account as a
    dictionary."""
    section = config[f"user{test_user_id}"]
    return {
        "email": section["email"],
        "name": section["name"],
        "affiliation": section["affiliation"],
        "password": section["password"]
    }


# get all users
def all_users():
    """Returns a dictionary containing User models,
    the keys are identifiers in the .ini file."""
    users = [get_user(id) for id in [1, 2, 3]]
    return {u.identifier: u for u in users}



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests


class OAuthHandler:
    def __init__(self, configs=None):
        if not (bool(configs) and isinstance(configs, dict)):
            raise ValueError("OAuthHandler needs a configuration dictionary.")

        # check if all necessary config parameters are there
        services = {}
        for provider, config in configs.items():
            # get required parameters
            authorization_url = config.get("AUTHORIZATION_URL", False)
            token_url = config.get("TOKEN_URL", False)
            client_id = config.get("CLIENT_ID", False)
            secret = config.get("CLIENT_SECRET", False)
            scope = config.get("SCOPE", "")
            if not (
                all(
                    [
                        bool(authorization_url),
                        bool(token_url),
                        bool(client_id),
                        bool(secret),
                    ]
                )
            ):
                raise ValueError(f"OAuthHandler has insufficient data for {provider}")
            else:
                # rebuild config
                services[provider.lower()] = {
                    "authorization_url": authorization_url,
                    "token_url": token_url,
                    "client_id": client_id,
                    "secret": secret,
                    "scope": scope,
                }
        # set the config dictionary
        self.services = services

    def front_end_params(self):
        """prepare a service dictionary for front-end: remove secrets"""
        result = {}
        # remove secret from service parameters
        for k, v in self.services.items():
            c = v.copy()
            c.pop("secret")
            result[k] = c
        return result

    def providers(self):
        """Returns a list with stored providers"""
        return list(self.services.keys())

    def get_user_credentials(self, provider, code, redirect_uri=None):
        """Extract User credentials with the help of a code"""
        result = False
        if provider == "github":
            result = self.__handle_github(code)
        elif provider == "orcid":
            result = self.__handle_orcid(code)
        elif provider == "google":
            result = self.__handle_google(code, redirect_uri)
        else:
            raise ValueError(f"Could not find provider {provider}")
        return result

    def __handle_orcid(self, code):
        """Handles OAuth roundtrip for Orcid"""
        # request token
        params = self.services["orcid"]
        response = requests.post(
            params["token_url"],
            data={
                "code": code,
                "client_id": params["client_id"],
                "client_secret": params["secret"],
                "grant_type": "authorization_code",
                "scope": "/authenticate",
            },
            headers={"Accept": "application/json"},
        ).json()
        id = response["orcid"]
        name = response.get("name", "")
        # TODO@Casper: I don't understand why I can't
        # get an email address from the public API. The
        # next call responds with a 401 Unauthorized which
        # doesn't make sense (I've set my email address on 'public')
        # because the call and the scope should OK. Why!?
        # token = response['access_token']
        # response = requests.get(
        #     f'https://api.sandbox.orcid.org/v3.0/{id}/email',
        #     headers={
        #         'Authorization': f'Bearer {token}',
        #         'Accept': 'application/json'
        #     }
        # ).json()
        # print(response.json())
        return (id, "email-unknown", name)

    def __handle_github(self, code):
        """Handles OAuth roundtrip for GitHub"""
        # request token
        params = self.services["github"]
        response = requests.post(
            params["token_url"],
            data={
                "code": code,
                "client_id": params["client_id"],
                "client_secret": params["secret"],
            },
            headers={"Accept": "application/json"},
        ).json()
        # if all is well, we hava a token
        token = response.get("access_token", "")
        # get a user profile
        response = requests.get(
            "https://api.github.com/user",
            headers={"Authorization": f"Bearer {token}", "Accept": "application/json"},
        )
        response = response.json()
        id = response["id"]
        email = response.get("email", "email-unknown")
        name = response["name"] or response["login"] or response["id"]
        return (id, email, name)

    def __handle_google(self, code, redirect_uri):
        # request token
        params = self.services["google"]
        response = requests.post(
            params["token_url"],
            data={
                "code": code,
                "client_id": params["client_id"],
                "client_secret": params["secret"],
                "grant_type": "authorization_code",
                "redirect_uri": redirect_uri,
            },
            headers={"Accept": "application/json"},
        ).json()
        # if all is well, we hava a token
        token = response.get("access_token", "")
        # get email
        response = requests.post(
            "https://www.googleapis.com/oauth2/v3/userinfo",
            data={"access_token": token},
            headers={"Accept": "application/json"},
        ).json()
        id = response["sub"]
        email = response.get("email", "email-unknown")
        name = (
            response.get("name", False) or response.get("family_name", False) or "Name"
        )
        return (id, email, name)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["ReviewSimulate"]

from datetime import datetime

import numpy as np
import pandas as pd

from asreview.project import open_state
from asreview.review import BaseReview
from asreview.review.base import LABEL_NA
from asreview.utils import get_random_state


def sample_prior_knowledge(
    labels, n_prior_included=10, n_prior_excluded=10, random_state=None
):
    """Function to sample prelabelled articles.

    Arguments
    ---------
    labels: np.ndarray
        Array of labels, with 1 -> included, 0 -> excluded.
    n_prior_included: int
        The number of positive labels.
    n_prior_excluded: int
        The number of negative labels.
    random_state : int, asreview.utils.SeededRandomState instance or None,
        optional (default=None)
        Random state or it's seed.

    Returns
    -------
    np.ndarray:
        An array with n_included and n_excluded indices.

    """
    # set random state
    r = get_random_state(random_state)

    # retrieve the index of included and excluded papers
    included_idx = np.where(labels == 1)[0]
    excluded_idx = np.where(labels == 0)[0]

    if len(included_idx) < n_prior_included:
        raise ValueError(
            f"Number of included priors requested ({n_prior_included})"
            f" is bigger than number of included papers "
            f"({len(included_idx)})."
        )
    if len(excluded_idx) < n_prior_excluded:
        raise ValueError(
            f"Number of excluded priors requested ({n_prior_excluded})"
            f" is bigger than number of excluded papers "
            f"({len(excluded_idx)})."
        )
    # select randomly from included and excluded papers
    included_indexes_sample = r.choice(included_idx, n_prior_included, replace=False)
    excluded_indexes_sample = r.choice(excluded_idx, n_prior_excluded, replace=False)

    init = np.append(included_indexes_sample, excluded_indexes_sample)

    return init


def naive_prior_knowledge(labels):
    """Select top records until the first 0 and 1 are found.

    Arguments
    ---------
    labels: np.ndarray
        Array of labels, with 1 -> included, 0 -> excluded.

    Returns
    -------
    np.ndarray:
        An array with prior indices from top dataset.

    """

    # retrieve the index of included and excluded papers
    first_included_idx = np.where(labels == 1)[0].min()
    first_excluded_idx = np.where(labels == 0)[0].min()

    return np.arange(max(first_included_idx, first_excluded_idx) + 1)


class ReviewSimulate(BaseReview):
    """ASReview Simulation mode class.

    Arguments
    ---------
    as_data: asreview.ASReviewData
        The data object which contains the text, labels, etc.
    model: BaseModel
        Initialized model to fit the data during active learning.
        See asreview.models.utils.py for possible models.
    query_model: BaseQueryModel
        Initialized model to query new instances for review, such as random
        sampling or max sampling.
        See asreview.query_strategies.utils.py for query models.
    balance_model: BaseBalanceModel
        Initialized model to redistribute the training data during the
        active learning process. They might either resample or undersample
        specific papers.
    feature_model: BaseFeatureModel
        Feature extraction model that converts texts and keywords to
        feature matrices.
    n_prior_included: int
        Sample n prior included papers.
    n_prior_excluded: int
        Sample n prior excluded papers.
    prior_indices: int
        Prior indices by row number.
    n_instances: int
        Number of papers to query at each step in the active learning
        process.
    stop_if: int
        Number of steps/queries to perform. Set to None for no limit.
    start_idx: numpy.ndarray
        Start the simulation/review with these indices. They are assumed to
        be already labeled. Failing to do so might result bad behaviour.
    init_seed: int
        Seed for setting the prior indices if the --prior_idx option is
        not used. If the option prior_idx is used with one or more
        index, this option is ignored.
    state_file: str
        Path to state file.
    write_interval: int
        After how many labeled records to write the simulation data to the
        state.
    """

    name = "simulate"

    def __init__(
        self,
        as_data,
        *args,
        n_prior_included=0,
        n_prior_excluded=0,
        prior_indices=None,
        init_seed=None,
        write_interval=None,
        **kwargs,
    ):
        self.n_prior_included = n_prior_included
        self.n_prior_excluded = n_prior_excluded

        self.write_interval = write_interval

        # check for partly labeled data
        labels = as_data.labels
        labeled_idx = np.where((labels == 0) | (labels == 1))[0]
        if len(labeled_idx) != len(labels):
            raise ValueError("Expected fully labeled dataset.")

        if prior_indices is not None and len(prior_indices) != 0:
            start_idx = prior_indices
        else:
            start_idx = as_data.prior_data_idx
            if len(start_idx) == 0 and n_prior_included + n_prior_excluded > 0:
                start_idx = sample_prior_knowledge(
                    labels, n_prior_included, n_prior_excluded, random_state=init_seed
                )
            else:
                start_idx = naive_prior_knowledge(labels)

        super(ReviewSimulate, self).__init__(
            as_data, *args, start_idx=start_idx, **kwargs
        )

        # Setup the reviewer attributes that take over the role of state
        # functions.
        with open_state(self.project) as state:
            # Check if there is already a ranking stored in the state.
            if state.model_has_trained:
                self.last_ranking = state.get_last_ranking()
                self.last_probabilities = state.get_last_probabilities()
            else:
                self.last_ranking = None
                self.last_probabilities = None

            self.labeled = state.get_labeled()
            self.pool = pd.Series(
                [
                    record_id
                    for record_id in self.record_table
                    if record_id not in self.labeled["record_id"].values
                ]
            )
            self.training_set = len(self.labeled)

            # Get the number of queries.
            training_sets = state.get_training_sets()
            # There is one query per trained model. We subtract 1
            # for the priors.
            self.total_queries = len(set(training_sets)) - 1

            # Check that both labels are available.
            if (0 not in self.labeled["label"].values) or (
                1 not in self.labeled["label"].values
            ):
                raise ValueError(
                    "Not both labels available Make sure there"
                    " is an included and excluded record in "
                    "the priors."
                )

        self.results = pd.DataFrame(
            [],
            columns=[
                "record_id",
                "label",
                "classifier",
                "query_strategy",
                "balance_strategy",
                "feature_extraction",
                "training_set",
                "labeling_time",
                "notes",
            ],
        )

    def _label_priors(self):
        """Make sure all the priors are labeled as well as the pending
        labels."""
        with open_state(self.project, read_only=False) as state:
            # Make sure the prior records are labeled.
            labeled = state.get_labeled()
            unlabeled_priors = [
                x for x in self.prior_indices if x not in labeled["record_id"].to_list()
            ]
            labels = self.data_labels[unlabeled_priors]

            with open_state(self.project, read_only=False) as s:
                s.add_labeling_data(unlabeled_priors, labels, prior=True)

            # Make sure the pending records are labeled.
            pending = state.get_pending()
            pending_labels = self.data_labels[pending]
            state.add_labeling_data(pending, pending_labels)

    def review(self):
        super().review()
        self._write_to_state()

    def _stop_review(self):
        """In simulation mode, the stop review function should get the labeled
        records list from the reviewer attribute."""

        # if the pool is empty, always stop
        if self.pool.empty:
            return True

        # If stop_if is set to min, stop when all papers in the pool are
        # irrelevant.
        if self.stop_if == "min" and (self.data_labels[self.pool] == 0).all():
            return True

        # Stop when reaching stop_if (if provided)
        if isinstance(self.stop_if, int) and self.total_queries >= self.stop_if:
            return True

        return False

    def train(self):
        """Train a new model on the labeled data."""
        # Check if both labels are available is done in init for simulation.
        # Use the balance model to sample the trainings data.
        new_training_set = len(self.labeled)

        y_sample_input = (
            pd.DataFrame(self.record_table)
            .merge(self.labeled, how="left", on="record_id")
            .loc[:, "label"]
            .fillna(LABEL_NA)
            .to_numpy()
        )
        train_idx = np.where(y_sample_input != LABEL_NA)[0]

        X_train, y_train = self.balance_model.sample(self.X, y_sample_input, train_idx)

        # Fit the classifier on the trainings data.
        self.classifier.fit(X_train, y_train)

        # Use the query strategy to produce a ranking.
        ranked_record_ids, relevance_scores = self.query_strategy.query(
            self.X, classifier=self.classifier, return_classifier_scores=True
        )

        self.last_ranking = pd.concat(
            [pd.Series(ranked_record_ids), pd.Series(range(len(ranked_record_ids)))],
            axis=1,
        )
        self.last_ranking.columns = ["record_id", "label"]
        # The scores for the included records in the second column.
        self.last_probabilities = relevance_scores[:, 1]

        self.training_set = new_training_set

    def _query(self, n):
        """In simulation mode, the query function should get the n highest
        ranked unlabeled records, without writing the model data to the results
        table. The"""
        unlabeled_ranking = self.last_ranking[
            self.last_ranking["record_id"].isin(self.pool)
        ]

        self.total_queries += 1

        return unlabeled_ranking["record_id"].iloc[:n].to_list()

    def _label(self, record_ids, prior=False):
        """In simulation mode, the label function should also add the model
        data to the results table."""

        labels = self.data_labels[record_ids]
        labeling_time = datetime.now()

        results = []
        for record_id, label in zip(record_ids, labels):
            results.append(
                {
                    "record_id": int(record_id),
                    "label": int(label),
                    "classifier": self.classifier.name,
                    "query_strategy": self.query_strategy.name,
                    "balance_strategy": self.balance_model.name,
                    "feature_extraction": self.feature_extraction.name,
                    "training_set": int(self.training_set),
                    "labeling_time": str(labeling_time),
                    "notes": None,
                }
            )

        self.results = pd.concat(
            [self.results, pd.DataFrame(results)], ignore_index=True
        )

        # Add the record ids to the labeled and remove from the pool.
        new_labeled_data = pd.DataFrame(
            zip(record_ids, labels), columns=["record_id", "label"]
        )
        self.labeled = pd.concat([self.labeled, new_labeled_data], ignore_index=True)
        self.pool = self.pool[~self.pool.isin(record_ids)]

        if (self.write_interval is not None) and (
            len(self.results) >= self.write_interval
        ):
            self._write_to_state()

        return labels

    def _write_to_state(self):
        """Write the data that has not yet been written to the state."""
        # Write the data to the state.
        if len(self.results) > 0:
            rows = [tuple(self.results.iloc[i]) for i in range(len(self.results))]
            with open_state(self.project, read_only=False) as state:
                state._add_labeling_data_simulation_mode(rows)

                state.add_last_ranking(
                    self.last_ranking["record_id"].to_numpy(),
                    self.classifier.name,
                    self.query_strategy.name,
                    self.balance_model.name,
                    self.feature_extraction.name,
                    self.training_set,
                )
                state.add_last_probabilities(self.last_probabilities)

            # Empty the results table in memory.
            self.results.drop(self.results.index, inplace=True)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging
import os
import socket
import time
import webbrowser
from threading import Timer

import requests
from gevent.pywsgi import WSGIServer

from asreview import __version__
from asreview._deprecated import DeprecateAction
from asreview._deprecated import mark_deprecated_help_strings
from asreview.project import ASReviewProject
from asreview.project import get_project_path
from asreview.project import get_projects
from asreview.webapp.app import create_app

# Host name
HOST_NAME = os.getenv("ASREVIEW_HOST")
if HOST_NAME is None:
    HOST_NAME = "localhost"

PORT_NUMBER = 5000


def _deprecated_dev_mode():
    if os.environ.get("FLASK_DEBUG", "") == "1":
        print(
            "\n\n\n!IMPORTANT!\n\n"
            "asreview lab development mode is deprecated, see:\n"
            "https://github.com/J535D165/asreview/blob/master/DEVELOPMENT.md"
            "\n\n\n"
        )
        exit(1)


def _check_port_in_use(host, port):
    logging.info(f"Checking if host and port are available :: {host}:{port}")
    host = host.replace("https://", "").replace("http://", "")
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def _open_browser(start_url):
    Timer(1, lambda: webbrowser.open_new(start_url)).start()

    print("\n\n\n\nIf your browser doesn't open. " f"Navigate to {start_url}\n\n\n\n")


def _check_for_update():
    """Check if there is an update available."""

    try:
        r = requests.get("https://pypi.org/pypi/asreview/json")
        r.raise_for_status()
        latest_version = r.json()["info"]["version"]
        if latest_version != __version__ and "+" not in __version__:
            print(
                "\n\n\n"
                f"ASReview LAB version {latest_version} is available. "
                "Please update using:\n"
                "pip install --upgrade asreview"
                "\n\n\n"
            )

            time.sleep(5)
    except Exception:
        print("Could not check for updates.")


def lab_entry_point(argv):
    # check deprecated dev mode
    _deprecated_dev_mode()

    parser = _lab_parser()
    mark_deprecated_help_strings(parser)
    args = parser.parse_args(argv)

    # check for update
    if not args.skip_update_check:
        _check_for_update()

    app = create_app(
        env="production",
        config_file=args.flask_config_file,
        secret_key=args.secret_key,
        salt=args.salt,
        enable_authentication=args.enable_authentication,
    )
    app.config["PROPAGATE_EXCEPTIONS"] = False

    # clean all projects
    # TODO@{Casper}: this needs a little bit
    # of work, we need to access all sub-folders
    if args.clean_all_projects:
        print("Cleaning all project files.")
        for project in get_projects():
            project.clean_tmp_files()
        print("Done")
        return

    # clean project by project_id
    # TODO@{Casper}: cleaning without a user context
    # is meaningless -> I think we should remove this
    # option
    if args.clean_project is not None:
        print(f"Cleaning project file '{args.clean_project}'.")
        ASReviewProject(get_project_path(args.clean_project)).clean_tmp_files()
        print("Done")
        return

    # if port is already taken find another one
    port = args.port
    original_port = port
    while _check_port_in_use(args.host, port) is True:
        old_port = port
        port = int(port) + 1
        if port - original_port >= args.port_retries:
            raise ConnectionError(
                "Could not find an available port \n"
                "to launch ASReview LAB. Last port \n"
                f"was {str(port)}"
            )
        print(f"Port {old_port} is in use.\n* Trying to start at {port}")

    protocol = "https://" if args.certfile and args.keyfile else "http://"
    start_url = f"{protocol}{args.host}:{port}/"

    ssl_args = {}
    if args.keyfile and args.certfile:
        ssl_args = {"keyfile": args.keyfile, "certfile": args.certfile}

    server = WSGIServer((args.host, port), app, **ssl_args)
    print(f"Serving ASReview LAB at {start_url}")

    if not args.no_browser:
        _open_browser(start_url)

    try:
        server.serve_forever()
    except KeyboardInterrupt:
        print("\n\nShutting down server\n\n")


def _lab_parser():
    # parse arguments if available
    parser = argparse.ArgumentParser(
        prog="lab",
        description="""ASReview LAB - Active learning for Systematic Reviews.""",  # noqa
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--clean-project",
        dest="clean_project",
        default=None,
        type=str,
        help="Safe cleanup of temporary files in project.",
    )

    parser.add_argument(
        "--clean-all-projects",
        dest="clean_all_projects",
        default=None,
        action="store_true",
        help="Safe cleanup of temporary files in all projects.",
    )

    parser.add_argument(
        "--ip",
        default=HOST_NAME,
        type=str,
        action=DeprecateAction,
        help="The IP address the server will listen on. Use the --host argument.",
    )

    parser.add_argument(
        "--host",
        default=HOST_NAME,
        type=str,
        help="The IP address the server will listen on.",
    )

    parser.add_argument(
        "--port",
        default=PORT_NUMBER,
        type=int,
        help="The port the server will listen on.",
    )

    parser.add_argument(
        "--enable-auth",
        dest="enable_authentication",
        action="store_true",
        help="Enable authentication.",
    )

    parser.add_argument(
        "--secret-key",
        default=None,
        type=str,
        help="Secret key for authentication.",
    )

    parser.add_argument(
        "--salt",
        default=None,
        type=str,
        help="When using authentication, a salt code is needed for hasing passwords.",
    )

    parser.add_argument(
        "--flask-configfile",
        dest="flask_config_file",
        type=str,
        help="Full path to a TOML file containing Flask parameters"
        "for authentication.",
    )

    parser.add_argument(
        "--no-browser",
        dest="no_browser",
        action="store_true",
        help="Do not open ASReview LAB in a browser after startup.",
    )

    parser.add_argument(
        "--port-retries",
        dest="port_retries",
        default=50,
        type=int,
        help="The number of additional ports to try if the"
        "specified port is not available.",
    )

    parser.add_argument(
        "--certfile",
        default="",
        type=str,
        help="The full path to an SSL/TLS certificate file.",
    )

    parser.add_argument(
        "--keyfile",
        default="",
        type=str,
        help="The full path to a private key file for usage with SSL/TLS.",
    )

    parser.add_argument(
        "--skip-update-check",
        dest="skip_update_check",
        action="store_true",
        help="Skip checking for updates.",
    )

    return parser



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
from pathlib import Path

from filelock import FileLock
from filelock import Timeout

from asreview.models.balance import get_balance_model
from asreview.models.classifiers import get_classifier
from asreview.models.feature_extraction import get_feature_model
from asreview.models.query import get_query_model
from asreview.project import ASReviewProject
from asreview.project import open_state
from asreview.review.base import BaseReview
from asreview.webapp.io import read_data


def run_model_entry_point(argv):

    # parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("project_path", type=str, help="Project id")
    parser.add_argument(
        "--output_error",
        dest="output_error",
        action="store_true",
        help="Save training error message to file.",
    )
    args = parser.parse_args(argv)

    project = ASReviewProject(args.project_path)

    try:

        # Check if there are new labeled records to train with
        with open_state(project.project_path) as state:
            if not state.exist_new_labeled_records:
                return

        # Lock so that only one training run is running at the same time.
        lock = FileLock(Path(project.project_path, "training.lock"), timeout=0)

        with lock:

            with open_state(project) as state:
                settings = state.settings

            reviewer = BaseReview(
                read_data(project),
                project,
                model=get_classifier(settings.model),
                query_model=get_query_model(settings.query_strategy),
                balance_model=get_balance_model(settings.balance_strategy),
                feature_model=get_feature_model(settings.feature_extraction),
                # random_state = get_random_state(seed)  # todo
            )

            reviewer.train()

        project.update_review(status="review")

    except Timeout:
        logging.debug("Another iteration is training")

    except Exception as err:
        project.set_error(err, save_error_message=args.output_error)
        raise err

    else:
        project.update_review(status="review")



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["RISReader"]

import io
import logging
import re
from urllib.request import urlopen

import pandas
import rispy

from asreview.io.utils import _standardize_dataframe
from asreview.utils import is_url


class RISReader:
    """RIS file reader."""

    read_format = [".ris", ".txt"]
    write_format = [".csv", ".tsv", ".xlsx", ".ris"]

    def _strip_zotero_p_tags(note_list):
        """Converter function for removing the XHTML <p></p> tags from Zotero export.

        Arguments
        ---------
        note_list: list
            A list of notes, coming from the Dataframe's "notes" column.

        Returns
        -------
        new_notes: list
            A list of updated notes, where XHTML <p></p> tags have been stripped.
        note_list: list
            The original note_list, when no XHTML <p></p> tags have been found.
        """
        if isinstance(note_list, list):
            new_notes = []
            for v in note_list:
                try:
                    new_notes.append(re.sub(r"^<p>|<\/p>$", "", v))
                except Exception:
                    new_notes.append(v)
            return new_notes
        else:
            return note_list

    def _label_parser(note_list):
        """Parse "included" and "notes" columns.

        Arguments
        ---------
        note_list: list
            A list of notes, coming from the Dataframe's "notes" column.

        Returns
        -------
        asreview_new_notes: list
            A list of updated notes, where internal label has been added.
        note_list: list
            The original note_list, when no labels have been found.
        1,0,-1: int
            Labels in case they are still needed from the internal representation.
        """
        regex = r"ASReview_relevant|ASReview_irrelevant|ASReview_not_seen"

        # Check whether note_list is actually a list and not NaN
        # Return -1 and an empty list
        if not isinstance(note_list, list):
            return -1, []

        # Create lists of lists for ASReview references
        asreview_refs = [re.findall(regex, note) for note in note_list]
        asreview_refs_list = [item for sublist in asreview_refs for item in sublist]

        if len(asreview_refs_list) > 0:
            # Create lists of lists for notes without references
            asreview_new_notes = [re.sub(regex, "", note) for note in note_list]
            # Remove empty elements from list
            asreview_new_notes[:] = [item for item in asreview_new_notes if item != ""]
            label = asreview_refs_list[-1]

            # Check for the label and return proper values for internal representation
            if label == "ASReview_relevant":
                return 1, asreview_new_notes
            elif label == "ASReview_irrelevant":
                return 0, asreview_new_notes
            elif label == "ASReview_not_seen":
                return -1, asreview_new_notes
        else:
            return -1, note_list

    @classmethod
    def read_data(cls, fp):
        """Import dataset.

        Arguments
        ---------
        fp: str, pathlib.Path
            File path to the RIS file.
        note_list: list
            A list of notes, coming from the Dataframe's "notes" column.

        Returns
        -------
        pandas.DataFrame:
            Dataframe with entries.

        Raises
        ------
        ValueError
            File with unrecognized encoding is used as input.
        """
        encodings = ["utf-8", "utf-8-sig", "ISO-8859-1"]
        entries = None
        if entries is None:
            if is_url(fp):
                url_input = urlopen(fp)
            for encoding in encodings:
                if is_url(fp):
                    try:
                        bibliography_file = io.StringIO(
                            url_input.read().decode(encoding)
                        )

                        entries = list(
                            rispy.load(bibliography_file, skip_unknown_tags=True)
                        )
                        bibliography_file.close()
                        break
                    except UnicodeDecodeError:
                        pass
                else:
                    try:
                        with open(fp, "r", encoding=encoding) as bibliography_file:
                            entries = list(
                                rispy.load(bibliography_file, skip_unknown_tags=True)
                            )
                            break
                    except UnicodeDecodeError:
                        pass
                    except IOError as e:
                        logging.warning(e)
            if entries is None:
                raise ValueError("Cannot find proper encoding for data file.")

        # Turn the entries dictionary into a Pandas dataframe
        df = pandas.DataFrame(entries)

        # Check if "notes" column is present
        if "notes" in df:
            # Strip Zotero XHTML <p> tags on "notes"
            df["notes"] = df["notes"].apply(cls._strip_zotero_p_tags)
            # Split "included" from "notes"
            df[["included", "notes"]] = pandas.DataFrame(
                df["notes"].apply(cls._label_parser).tolist(),
                columns=["included", "notes"],
            )
            # Return the standardised dataframe with label and notes separated
            return _standardize_dataframe(df)
        else:
            # Return the standardised dataframe
            return _standardize_dataframe(df)



from flask import current_app

import asreview.webapp.tests.utils.api_utils as au


# Test if index.html is available!
# Note: This test will fail if build is missing. Please run
# `python setup.py compile_assets` first.
def test_landing(setup):
    client, _, _ = setup

    status_code, _, html = au.call_root_url(client)
    assert status_code == 200
    assert (
        "<title>ASReview LAB - A tool for AI-assisted systematic reviews</title>"
        in html
    )  # noqa


# Test boot data!
def test_boot(setup_all_clients):
    status_code, data = au.call_boot_url(setup_all_clients)
    assert status_code == 200
    assert isinstance(data, dict)
    assert "authentication" in data.keys()
    # assert "status" in data.keys()  # what is the aim of this?
    assert "version" in data.keys()
    if not current_app.config.get("LOGIN_DISABLED"):
        assert data["authentication"]
        assert data["allow_account_creation"] == current_app.config.get(
            "ALLOW_ACCOUNT_CREATION"
        )
        assert data["email_verification"] == current_app.config.get(
            "EMAIL_VERIFICATION", False
        )
    else:
        assert not data["authentication"]



import random

import asreview.webapp.tests.integration_tests.utils as utils
from asreview.webapp.authentication.models import Project

ACCOUNT = {
    "email": "test4@user.org",
    "name": "Test User",
    "affiliation": "Utrecht University",
    "password": "@Secret1234!"
}

PROJECT = {
    "mode": "explore",
    "title": "Project Title",
    "author": ACCOUNT["name"],
    "description": "Project description",
    "dataset": {
        "type": "benchmark",
        "label": "Appenzeller‐Herzog et al. (2019)",

        # alternative:
        # "prior_knowledge_method": "Random",
        # "prior_knowledge": [
        #     "relevant",
        #     "irrelevant",
        #     "relevant",
        #     "irrelevant",
        #     "relevant"
        # ],

        "prior_knowledge_method": "Search",
        "prior_knowledge": [
            ("hepatolenticular degeneration Wilson Zinc", None),
            ("hepatolenticular degeneration Wilson ovulatory", None),
            ("triethylenetetramine dihydrochloride", None),
            ("disease genetic heterogeneity United Kingdom and Taiwan", None),
            ("AIM succimer penicillamine", None),
            ("Early neurological worsening", None)
        ],
    },

    "model": {
        "feature_extraction": "tfidf",
        "classifier": "nb",
        "query_strategy": "max",
        "balance_strategy": "double"
    }
}


def test_signup_signin_create_project(driver, url, database_uri, reading_time):
    base_url = url
    driver.get(base_url)

    # setup database session
    session = utils.setup_database_session(database_uri)

    # clean database
    utils.clean_database(session)

    # check if we have a no registered projects
    assert len(session.query(Project).all()) == 0

    # create account
    utils.create_account(driver, base_url, ACCOUNT)

    # sign in
    utils.sign_in(driver, base_url, ACCOUNT)

    # create project
    utils.create_project(driver, base_url, PROJECT, reading_time)

    # check if we have a registered project in the database
    assert len(session.query(Project).all()) == 1

    # REVIEWING
    for _ in range(5):
        # choose
        label = random.choice(['irrelevant', 'relevant'])
        # click
        utils.label_abstract(driver, label, reading_time)

    # close driver
    driver.close()



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["list_classifiers", "get_classifier_class", "get_classifier"]

import logging

from asreview.utils import _entry_points


def _set_class_weight(weight1):
    """Used various classifiers to have quicker learning."""
    if weight1 is None:
        return None
    weight0 = 1.0
    cw_class = {
        0: weight0,
        1: weight1,
    }
    logging.debug(f"Using class weights: 0 <- {weight0}, 1 <- {weight1}")
    return cw_class


def list_classifiers():
    """List available classifier classes.

    Returns
    -------
    list:
        Classes of available classifiers in alphabetical order.
    """

    return [e.load() for e in _entry_points(group="asreview.models.classifiers")]


def get_classifier_class(name):
    """Get class of model from string.

    Arguments
    ---------
    name: str
        Name of the model, e.g. 'svm', 'nb' or 'lstm-pool'.

    Returns
    -------
    BaseModel:
        Class corresponding to the name.
    """
    return _entry_points(group="asreview.models.classifiers")[name].load()


def get_classifier(name, *args, random_state=None, **kwargs):
    """Get an instance of a model from a string.

    Arguments
    ---------
    name: str
        Name of the model.
    *args:
        Arguments for the model.
    **kwargs:
        Keyword arguments for the model.

    Returns
    -------
    BaseFeatureExtraction:
        Initialized instance of classifier.
    """
    model_class = get_classifier_class(name)
    try:
        return model_class(*args, random_state=random_state, **kwargs)
    except TypeError:
        return model_class(*args, **kwargs)



import pytest
from sqlalchemy.exc import IntegrityError

import asreview.webapp.tests.utils.crud as crud
from asreview.webapp import DB
from asreview.webapp.authentication.models import Collaboration
from asreview.webapp.authentication.models import CollaborationInvitation


class TestInvitations:
    """Testing invitations on database level."""

    # #############
    # CREATE
    # #############

    # test adding an invitation
    def test_adding_an_invitation(self, test_data):
        # get project
        project = test_data["user1"].projects[0]
        # invite user 2
        project.pending_invitations.append(test_data["user2"])
        DB.session.commit()

        assert crud.count_invitations() == 1
        # get fresh object
        invite = crud.last_invitation()
        # asserts Invitations
        assert invite.project_id == project.id
        assert invite.user_id == test_data["user2"].id

    # test uniqueness of invitations
    def test_uniqueness_of_invitations(self, test_data):
        user1 = test_data["user1"]
        user2 = test_data["user2"]
        project = user1.projects[0]
        crud.create_invitation(DB, project, user2)
        assert crud.count_invitations() == 1

        # create identical invitation
        with pytest.raises(IntegrityError):
            crud.create_invitation(DB, project, user2)
        # if all is well, we can't add the same invitation
        assert crud.count_invitations() == 1

    # test missing user is not permitted
    def test_missing_user_in_invitation(self, test_data):
        project = test_data["user1"].projects[0]
        invite = CollaborationInvitation(project_id=project.id, user_id=None)
        DB.session.add(invite)
        with pytest.raises(IntegrityError):
            DB.session.commit()
        DB.session.rollback()
        assert crud.count_invitations() == 0

    # test missing project is not permitted
    def test_missing_project_in_invitation(self, test_data):
        user = test_data["user1"]
        invite = CollaborationInvitation(project_id=None, user_id=user.id)
        DB.session.add(invite)
        with pytest.raises(IntegrityError):
            DB.session.commit()
        DB.session.rollback()
        assert crud.count_invitations() == 0


class TestCollaborations:
    """Testing collaboration on database level."""

    # #############
    # CREATE
    # #############

    # test adding a collaboration
    def test_create_collaboration(self, test_data):
        # get project
        project = test_data["user1"].projects[0]
        # collaboration user 2
        project.collaborators.append(test_data["user2"])
        DB.session.commit()
        assert crud.count_collaborations() == 1
        # get fresh objects
        collab = crud.last_collaboration()
        # asserts collaboration
        assert collab.project_id == project.id
        assert collab.user_id == test_data["user2"].id

    # test uniqueness of collaboration
    def test_uniqueness_of_collaboration(self, test_data):
        user1 = test_data["user1"]
        user2 = test_data["user2"]
        project = user1.projects[0]
        crud.create_collaboration(DB, project, user2)
        assert crud.count_collaborations() == 1

        # create identical invitation
        with pytest.raises(IntegrityError):
            crud.create_collaboration(DB, project, user2)
        # if all is well, we can't add the same invitation
        assert crud.count_collaborations() == 1

    # test missing user is not permitted
    def test_missing_user_in_collaboration(self, test_data):
        project = test_data["user1"].projects[0]
        invite = Collaboration(project_id=project.id, user_id=None)
        DB.session.add(invite)
        with pytest.raises(IntegrityError):
            DB.session.commit()
        DB.session.rollback()
        assert crud.count_collaborations() == 0

    # test missing project is not permitted
    def test_missing_project_in_collaboration(self, test_data):
        user = test_data["user1"]
        invite = Collaboration(project_id=None, user_id=user.id)
        DB.session.add(invite)
        with pytest.raises(IntegrityError):
            DB.session.commit()
        DB.session.rollback()
        assert crud.count_collaborations() == 0



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["Tfidf"]

from sklearn.feature_extraction.text import TfidfVectorizer

from asreview.models.feature_extraction.base import BaseFeatureExtraction


class Tfidf(BaseFeatureExtraction):
    """TF-IDF feature extraction technique (``tfidf``).

    Use the standard TF-IDF (Term Frequency-Inverse Document Frequency) feature
    extraction technique from `SKLearn <https://scikit-learn.org/stable/modules/
    generated/sklearn.feature_extraction.text.TfidfVectorizer.html>`__. Gives a
    sparse matrix as output. Works well in combination with
    :class:`asreview.models.classifiers.NaiveBayesClassifier` and other fast
    training models (given that the features vectors are relatively wide).

    Arguments
    ---------
    ngram_max: int
        Can use up to ngrams up to ngram_max. For example in the case of
        ngram_max=2, monograms and bigrams could be used.
    stop_words: str
        When set to 'english', use stopwords. If set to None or 'none',
        do not use stop words.
    """

    name = "tfidf"
    label = "TF-IDF"

    def __init__(self, *args, ngram_max=1, stop_words="english", **kwargs):
        """Initialize tfidf class."""
        super(Tfidf, self).__init__(*args, **kwargs)
        self.ngram_max = ngram_max
        self.stop_words = stop_words
        if stop_words is None or stop_words.lower() == "none":
            sklearn_stop_words = None
        else:
            sklearn_stop_words = self.stop_words
        self._model = TfidfVectorizer(
            ngram_range=(1, ngram_max), stop_words=sklearn_stop_words
        )

    def fit(self, texts):
        self._model.fit(texts)

    def transform(self, texts):
        X = self._model.transform(texts).tocsr()
        return X

    def full_hyper_space(self):
        from hyperopt import hp

        hyper_space, hyper_choices = super(Tfidf, self).full_hyper_space()
        hyper_choices.update({"fex_stop_words": ["english", "none"]})
        hyper_space.update(
            {
                "fex_ngram_max": hp.uniformint("fex_ngram_max", 1, 3),
                "fex_stop_words": hp.choice(
                    "fex_stop_words", hyper_choices["fex_stop_words"]
                ),
            }
        )
        return hyper_space, hyper_choices



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import warnings
from pathlib import Path

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from flask import Flask
from flask import send_from_directory
from flask.json import jsonify
from flask.templating import render_template
from flask_cors import CORS
from flask_login import LoginManager
from werkzeug.exceptions import InternalServerError

from asreview import __version__ as asreview_version
from asreview.utils import asreview_path
from asreview.webapp import DB
from asreview.webapp.api import auth
from asreview.webapp.api import projects
from asreview.webapp.api import team
from asreview.webapp.authentication.models import User
from asreview.webapp.authentication.oauth_handler import OAuthHandler


def create_app(
    env="development",
    config_file=None,
    secret_key=None,
    salt=None,
    enable_authentication=False,
):
    app = Flask(
        __name__,
        instance_relative_config=True,
        static_folder="build/static",
        template_folder="build",
    )

    app.config["ALLOWED_ORIGINS"] = ["http://localhost:3000", "http://127.0.0.1:3000"]

    app.config["SECRET_KEY"] = secret_key
    app.config["SALT"] = salt

    app.config.from_prefixed_env()

    # If allowed origins comes as an environment variables, and is a list,
    # then we assume a JSON string.
    if isinstance(app.config.get("ALLOWED_ORIGINS", False), str):
        app.config["ALLOWED_ORIGINS"] = json.loads(app.config["ALLOWED_ORIGINS"])

    if config_file_path := config_file or app.config.get("CONFIGFILE", ""):
        app.config.from_file(
            Path(config_file_path).absolute(), load=tomllib.load, text=False
        )

    if app.config.get("AUTHENTICATION_ENABLED", None):
        warnings.warn(
            "The use of AUTHENTICATION_ENABLED=true is deprecated and "
            "will be removed in the future. Use LOGIN_DISABLED=false instead."
        )
        if "LOGIN_DISABLED" not in app.config:
            app.config["LOGIN_DISABLED"] = False

    if "LOGIN_DISABLED" not in app.config:
        app.config["LOGIN_DISABLED"] = not enable_authentication

    if origins := app.config.get("ALLOWED_ORIGINS", False):
        CORS(app, origins=origins, supports_credentials=True)

    with app.app_context():
        app.register_blueprint(projects.bp)

    if not app.config.get("LOGIN_DISABLED", False):
        # config JSON Web Tokens
        login_manager = LoginManager(app)
        login_manager.init_app(app)
        login_manager.session_protection = "strong"

        # Register a callback function for current_user.
        @login_manager.user_loader
        def load_user(user_id):
            return User.query.get(int(user_id))

        if not app.config.get("SQLALCHEMY_DATABASE_URI", None):
            uri = os.path.join(asreview_path(), f"asreview.{env}.sqlite")
            app.config["SQLALCHEMY_DATABASE_URI"] = f"sqlite:///{uri}"

        # initialize app for SQLAlchemy
        DB.init_app(app)

        with app.app_context():
            # create tables in case they don't exist
            DB.create_all()

        # store oauth config in oauth handler
        if bool(app.config.get("OAUTH", False)):
            app.config["OAUTH"] = OAuthHandler(app.config["OAUTH"])

        with app.app_context():
            app.register_blueprint(auth.bp)
            app.register_blueprint(team.bp)

    # Ensure the instance folder exists.
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    @app.errorhandler(InternalServerError)
    def error_500(e):
        original = getattr(e, "original_exception", None)

        if original is None:
            # direct 500 error, such as abort(500)
            logging.error(e)
            return jsonify(message="Whoops, something went wrong."), 500

        # wrapped unhandled error
        logging.error(e.original_exception)
        return jsonify(message=str(e.original_exception)), 500

    @app.route("/", methods=["GET"])
    @app.route("/signin", methods=["GET"])
    @app.route("/signup", methods=["GET"])
    @app.route("/confirm_account", methods=["GET"])
    @app.route("/oauth_callback", methods=["GET"])
    @app.route("/get_profile", methods=["GET"])
    @app.route("/update_profile", methods=["GET"])
    @app.route("/projects/", methods=["GET"])
    @app.route("/projects/<project_id>/", methods=["GET"])
    @app.route("/projects/<project_id>/<tab>/", methods=["GET"])
    @app.route("/forgot_password", methods=["GET"])
    @app.route("/reset_password", methods=["GET"])
    def index():
        return render_template("index.html")

    @app.route("/favicon.ico")
    def send_favicon():
        return send_from_directory(
            "build", "favicon.ico", mimetype="image/vnd.microsoft.icon"
        )

    @app.route("/boot", methods=["GET"])
    def api_boot():
        """Get the boot info."""

        authenticated = not app.config.get("LOGIN_DISABLED", False)

        response = {
            "authentication": authenticated,
            "version": asreview_version,
        }

        if authenticated:
            # if recaptcha config is provided for account creation
            if app.config.get("RE_CAPTCHA_V3", False):
                response["recaptchav3_key"] = app.config["RE_CAPTCHA_V3"].get(
                    "KEY", False
                )

            response["allow_account_creation"] = app.config.get(
                "ALLOW_ACCOUNT_CREATION", False
            )
            response["allow_teams"] = app.config.get("ALLOW_TEAMS", False)

            response["email_verification"] = bool(
                app.config.get("EMAIL_VERIFICATION", False)
            )

            response["email_config"] = all([
                app.config.get("MAIL_SERVER", False),
                app.config.get("MAIL_USERNAME", False),
                app.config.get("MAIL_PASSWORD", False)
            ])

            # if oauth config is provided
            if isinstance(app.config.get("OAUTH", False), OAuthHandler):
                if params := app.config.get("OAUTH").front_end_params():
                    response["oauth"] = params

        return jsonify(response)

    return app



from pathlib import Path

from pytest import mark

from asreview import ASReviewData


@mark.parametrize("test_file", [("baseline_tag-notes_labels.ris")])
def test_asreview_labels_ris(test_file, tmpdir):
    fp_in = Path("tests", "demo_data", test_file)
    asr_data = ASReviewData.from_file(fp_in)

    tmp_ris_fp_out = Path(tmpdir, "tmp_labels.ris")
    asr_data.to_file(tmp_ris_fp_out)
    asr_data_diff = ASReviewData.from_file(tmp_ris_fp_out)

    # Check if input file matches the export file
    assert list(asr_data.title) == list(asr_data_diff.title)
    assert list(asr_data.labels) == list(asr_data_diff.labels)


@mark.parametrize("test_file", [("baseline_tag-notes.ris")])
def test_asreview_notes_ris(test_file, tmpdir):
    fp_in = Path("tests", "demo_data", test_file)
    asr_data = ASReviewData.from_file(fp_in)

    tmp_ris_fp_out = Path(tmpdir, "tmp_notes.ris")
    asr_data.to_file(tmp_ris_fp_out)

    asr_data_diff = ASReviewData.from_file(tmp_ris_fp_out)

    # Check if input file matches the export file
    assert list(asr_data.title) == list(asr_data_diff.title)
    assert list(asr_data.notes) == list(asr_data_diff.notes)


@mark.parametrize("test_file", [("ris_issue_992.txt"), ("ris_issue_1099.txt")])
def test_asreview_ris(test_file, tmpdir):
    fp_in = Path("tests", "demo_data", test_file)
    asr_data = ASReviewData.from_file(fp_in)

    tmp_ris_fp_out = Path(tmpdir, "tmp_ris.ris")
    asr_data.to_file(tmp_ris_fp_out)

    asr_data_diff = ASReviewData.from_file(tmp_ris_fp_out)

    # Check if input file matches the export file
    assert list(asr_data.title) == list(asr_data_diff.title)


def test_write_numpy_arrays():
    # This test should catch cases where two numpy arrays
    # are to be evaluated in boolean context. Error is as follows:
    # "The truth value of an array with more than one element is ambiguous.
    # Use a.any() or a.all()"
    ###
    # For ris writer, a relevant bug was fixed with commit 70d9497
    pass
